/**
 * Étoilé ProjectManager - XCBConnection.m
 *
 * Copyright (C) 2009 David Chisnall
 * Copyright (C) 2010 Christopher Armstrong <carmstrong@fastmail.com.au>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *
 **/
#import <XCBKit/XCBConnection.h>
#import <XCBKit/XCBScreen.h>
#import <XCBKit/XCBSelection.h>
#import "XCBWindow+Package.h"
#import <XCBKit/XCBAtomCache.h>
#import <XCBKit/XCBException.h>
#import <EtoileFoundation/EtoileFoundation.h>
#include <xcb/xcbext.h>
#include <xcb/damage.h>
#include <xcb/xcb_event.h>

@interface XCBConnection (EventHandlers)
- (void)handleMapNotify: (xcb_map_notify_event_t*)anEvent;
- (void)handleCreateNotify: (xcb_create_notify_event_t*)anEvent;
- (void)handleButtonPress: (xcb_button_press_event_t*)anEvent;
- (void)handleButtonRelease: (xcb_button_release_event_t*)anEvent;
- (void)handleKeyPress: (xcb_key_press_event_t*)anEvent;
- (void)handleKeyRelease: (xcb_key_release_event_t*)anEvent;
- (void)handleMotionNotify: (xcb_motion_notify_event_t*)anEvent;
- (void)handleEnterNotify: (xcb_enter_notify_event_t*)anEvent;
- (void)handleLeaveNotify: (xcb_leave_notify_event_t*)anEvent;
- (void)handleExpose: (xcb_expose_event_t*)anEvent;
- (void)handleMapRequest: (xcb_map_request_event_t*)anEvent;
- (void)handleCirculateRequest: (xcb_circulate_request_event_t*)anEvent;
- (void)handleConfigureRequest: (xcb_configure_request_event_t*)anEvent;
- (void)handleReparentNotify: (xcb_reparent_notify_event_t*)anEvent;
- (void)handlePropertyNotify: (xcb_property_notify_event_t*)anEvent;
@end

@interface XCBConnection (Private)
- (void)eventsReady: (NSNotification*)notification;
@end
#define HANDLE_EVENT(eventName, eventObj, eventWin, xcbWindow) \
	if (nil == xcbWindow) \
		NSLog(@ #eventName " without window (%x)", eventWin); \
	else \
		[xcbWindow handle##eventName : eventObj];

@implementation XCBConnection (EventHandlers)
- (void) handleFocusIn: (xcb_focus_in_event_t*)anEvent
{
	NSDebugLLog(@"XCBConnection", @"Focus in");
	XCBWindow *win = [self windowForXCBId: anEvent->event];
	HANDLE_EVENT(FocusIn, anEvent, anEvent->event, win)
	// NSAssert1(win, @"Focus in without window (%x)", anEvent->event);
	//[win handleFocusIn: anEvent];
}
- (void) handleFocusOut: (xcb_focus_out_event_t*)anEvent
{
	NSDebugLLog(@"XCBConnection", @"Focus out");
	XCBWindow *win = [self windowForXCBId: anEvent->event];
	//NSAssert1(win, @"Focus out without window (%x)", anEvent->event);
	//[win handleFocusOut: anEvent];
	HANDLE_EVENT(FocusOut, anEvent, anEvent->event, win);
}
- (void) handleMapRequest: (xcb_map_request_event_t*)anEvent
{
	NSDebugLLog(@"XCBConnection", @"Mapping requested for window: %x", anEvent->window);
	XCBWindow *win = [self windowForXCBId: anEvent->window];
	HANDLE_EVENT(MapRequest, anEvent, anEvent->window, win);
	//NSAssert1(win, @"Map request without window (%x)", anEvent->window);
	//[win handleMapRequest: anEvent];
}
- (void)handleCirculateRequest: (xcb_circulate_request_event_t*)anEvent
{
	XCBWindow *win = [self windowForXCBId: anEvent->window];
	HANDLE_EVENT(CirculateRequest, anEvent, anEvent->window, win);
	//NSAssert1(win, @"Circulate request without window (%x)", anEvent->window);
	//[win handleCirculateRequest: anEvent];
}
- (void)handleConfigureRequest: (xcb_configure_request_event_t*)anEvent
{
	NSDebugLLog(@"XCBConnection",@"XCBConnection: Configure requested for window: %x", anEvent->window);
	XCBWindow *win = [self windowForXCBId: anEvent->window];
	HANDLE_EVENT(ConfigureRequest, anEvent, anEvent->window, win);
	//NSAssert1(win, @"Configure request without window (%x)", anEvent->window);
	//[win handleConfigureRequest: anEvent];
}
- (void)handleButtonPress: (xcb_button_press_event_t*)anEvent
{
	NSDebugLLog(@"XCBConnection",@"Button pressed");
	XCBWindow *win = [self windowForXCBId: anEvent->event];
	//NSAssert1(win, @"Button Press without window (%x)", anEvent->event);
	currentTime = anEvent->time;
	HANDLE_EVENT(ButtonPress, anEvent, anEvent->event, win);
	//[win handleButtonPress: anEvent];
}
- (void)handleButtonRelease: (xcb_button_release_event_t*)anEvent
{
	NSDebugLLog(@"XCBConnection",@"Button released");
	XCBWindow *win = [self windowForXCBId: anEvent->event];
	//NSAssert1(win, @"Button Release without window (%x)", anEvent->event);
	currentTime = anEvent->time;
	HANDLE_EVENT(ButtonRelease, anEvent, anEvent->event, win);
	//[win handleButtonRelease: anEvent];
}
- (void)handleKeyPress: (xcb_key_press_event_t*)anEvent
{
	NSDebugLLog(@"XCBConnection",@"Key pressed");
	currentTime = anEvent->time;
}
- (void)handleKeyRelease: (xcb_key_release_event_t*)anEvent
{
	NSDebugLLog(@"XCBConnection",@"Key released");
	currentTime = anEvent->time;
}
- (void)handleMotionNotify: (xcb_motion_notify_event_t*)anEvent
{
	NSDebugLLog(@"XCBConnection",@"Motion notify");
	XCBWindow *win = [self windowForXCBId: anEvent->event];
	//NSAssert1(win, @"Motion Notify without window (%x)", anEvent->event);
	currentTime = anEvent->time;
	HANDLE_EVENT(MotionNotify, anEvent, anEvent->event, win);
	//[win handleMotionNotify: anEvent];
}
- (void)handleEnterNotify: (xcb_enter_notify_event_t*)anEvent
{
	NSDebugLLog(@"XCBConnection",@"Enter notify");
	currentTime = anEvent->time;
}
- (void)handleLeaveNotify: (xcb_leave_notify_event_t*)anEvent
{
	NSDebugLLog(@"XCBConnection",@"Leave notify");
	currentTime = anEvent->time;
}
- (void)handleConfigureNotify: (xcb_configure_notify_event_t*)anEvent
{
	XCBWindow *win = [self windowForXCBId: anEvent->window];
	//NSAssert1(win, @"Configure Notify without window (%x)", anEvent->window);
	//[win handleConfigureNotifyEvent: anEvent];
	HANDLE_EVENT(ConfigureNotifyEvent, anEvent, anEvent->window, win);
}
- (void)handleDestroyNotify: (xcb_destroy_notify_event_t*)anEvent
{
	XCBWindow *win = [self windowForXCBId: anEvent->window];
	//NSAssert1(win, @"Destroy notify without window (%x)", anEvent->window);
	HANDLE_EVENT(DestroyNotifyEvent, anEvent, anEvent->window, win);
	//[win handleDestroyNotifyEvent: anEvent];
}
- (void)handleUnMapNotify: (xcb_unmap_notify_event_t*)anEvent
{
	XCBWindow *win  = [self windowForXCBId: anEvent->window];
	//NSAssert1(win, @"Unmap Notify without window (%x)", anEvent->window);
	NSDebugLLog(@"XCBConnection",@"UnMapping window %@ (%x)", win, anEvent->window);
	//[win handleUnMapNotifyEvent: anEvent];
	HANDLE_EVENT(UnMapNotifyEvent, anEvent, anEvent->window, win);
}
- (void)handleMapNotify: (xcb_map_notify_event_t*)anEvent
{
	XCBWindow *win  = [self windowForXCBId: anEvent->window];
	//NSAssert1(win, @"Map Notify without window (%x)", anEvent->window);
	//[win handleMapNotifyEvent: anEvent];
	HANDLE_EVENT(MapNotifyEvent, anEvent, anEvent->window, win);
}
- (void)handleCirculateNotify: (xcb_circulate_notify_event_t*)anEvent
{
	XCBWindow *win  = [self windowForXCBId: anEvent->window];
	//NSAssert1(win, @"Circulate Notify without window (%x)", anEvent->window);
	//[win handleCirculateNotifyEvent: anEvent];
	HANDLE_EVENT(CirculateNotifyEvent, anEvent, anEvent->window, win);
}
- (void)handleCreateNotify: (xcb_create_notify_event_t*)anEvent
{
	NSDebugLLog(@"XCBConnection",@"Created window %x", anEvent->window);
	[XCBWindow windowWithCreateEvent: anEvent];
	// No need to post notification, as this is handled by XCBWindow itself.
}
- (void)handleExpose: (xcb_expose_event_t*)anEvent
{
	XCBWindow *win = [self windowForXCBId:anEvent->window];
	//NSAssert1(win, @"Expose without window (%x)", anEvent->window);
	//[win handleExpose: anEvent];
	HANDLE_EVENT(Expose, anEvent, anEvent->window, win);
}
- (void)handleReparentNotify: (xcb_reparent_notify_event_t*)anEvent
{
	XCBWindow *win = [self windowForXCBId: anEvent->window];
	//NSAssert1(win, @"Reparent without window (%x)", anEvent->window);
	//[win handleReparentNotify: anEvent];
	HANDLE_EVENT(ReparentNotify, anEvent, anEvent->window, win);
}
- (void)handlePropertyNotify: (xcb_property_notify_event_t*)anEvent
{
	XCBWindow *win = [self windowForXCBId: anEvent->window];
	//NSAssert1(win, @"Property notify without window (%x)", anEvent->window);
	//[win handlePropertyNotify: anEvent];
	HANDLE_EVENT(PropertyNotify, anEvent, anEvent->window, win);
}
@end

XCBConnection *XCBConn;


@implementation XCBConnection
- (void)startMessageLoop
{
	[self flush];
	[[NSRunLoop currentRunLoop] run];
}
+ (XCBConnection*)sharedConnection
{
	if (nil == XCBConn)
	{
		NSDebugLLog(@"XCBConnection",@"Creating shared connection...");
		[[self alloc] init];
	}
	return XCBConn;
}
- (id)init
{
	SUPERINIT;
	NSDebugLLog(@"XCBConnection",@"Creating connection...");
	NSDebugLLog(@"XCBConnection",@"Self: %p", self);
	[NSRunLoop currentRunLoop];
	connection = xcb_connect(NULL, NULL);
	if (NULL == connection)
	{
		[self release];
		return nil;
	}

	if (xcb_connection_has_error(connection)) 
	{
		NSLog(@"Unknown error creating connection.");
		[self release];
		return nil;
	}

	currentTime = XCB_CURRENT_TIME;

	replyHandlers = [NSMutableArray new];
	windows = NSCreateMapTable(NSIntMapKeyCallBacks,
			NSObjectMapValueCallBacks, 100);
	selections = NSCreateMapTable(NSIntMapKeyCallBacks,
			NSObjectMapValueCallBacks, 10);
	screens = [NSMutableArray new];

	// Hack needed because creating XCBWindow instances requires XCBConn to be
	// valid.
	XCBConn = self;

	// Set up event delivery
	int fd = xcb_get_file_descriptor(connection);
	if (-1 == fd)
	{
		NSLog(@"Received invalid file descriptor for XCBConnection");
		[self release];
		return nil;
	}
	handle = [[NSFileHandle alloc] initWithFileDescriptor:fd
	                                       closeOnDealloc:NO];
	NSNotificationCenter *center =
		[NSNotificationCenter defaultCenter];
	[center addObserver: self
	           selector: @selector(eventsReady:)
	               name: NSFileHandleDataAvailableNotification
	             object: handle];
	[handle waitForDataInBackgroundAndNotify];

	// Create XCBWindows for all of the root windows.
	xcb_screen_iterator_t iter = 
		xcb_setup_roots_iterator(xcb_get_setup(connection));
	while (iter.rem)
	{
		xcb_screen_t *screen = iter.data;
		[screens addObject: [XCBScreen screenWithXCBScreen: screen]];
		NSDebugLLog(@"XCBConnection",@"Root %x (%dx%d)", screen->root, screen->width_in_pixels, 
				screen->height_in_pixels);
		[self registerWindow: [XCBWindow windowWithXCBWindow: screen->root parent: XCB_NONE]];

		xcb_screen_next(&iter);
	}
	[self flush];
	NSDebugLLog(@"XCBConnection",@"Connection created");
	return self;
}

#define HANDLE(constant, sel) \
	case XCB_ ## constant:\
		NSDebugLLog(@"XCBConnection",@"Handling %s", #constant);\
		if ([self respondsToSelector:@selector(handle ## sel:)])\
		{\
			[self handle ## sel: (void*)event];\
		}\
		break;

- (BOOL)handleEvents
{
	BOOL eventsHandled = NO;
	NSDebugLLog(@"XCBConnection",@"Handling events");
	xcb_generic_event_t *event;
	while (NULL != (event = xcb_poll_for_event(connection)))
	{
		switch (event->response_type & XCB_EVENT_RESPONSE_TYPE_MASK)
		{
			HANDLE(KEY_PRESS, KeyPress)
			HANDLE(KEY_RELEASE, KeyRelease)
			HANDLE(BUTTON_RELEASE, ButtonRelease)
			HANDLE(MOTION_NOTIFY, MotionNotify)
			HANDLE(ENTER_NOTIFY, EnterNotify)
			HANDLE(LEAVE_NOTIFY, LeaveNotify)
			HANDLE(FOCUS_OUT, FocusOut)
			HANDLE(FOCUS_IN, FocusIn)
			HANDLE(BUTTON_PRESS, ButtonPress)
			HANDLE(MAP_REQUEST, MapRequest)
			HANDLE(CIRCULATE_REQUEST, CirculateRequest)
			HANDLE(CONFIGURE_REQUEST, ConfigureRequest)
			HANDLE(UNMAP_NOTIFY, UnMapNotify)
			HANDLE(MAP_NOTIFY, MapNotify)
			HANDLE(DESTROY_NOTIFY, DestroyNotify)
			HANDLE(CREATE_NOTIFY, CreateNotify)
			HANDLE(CONFIGURE_NOTIFY, ConfigureNotify)
			HANDLE(CIRCULATE_NOTIFY, CirculateNotify)
			HANDLE(REPARENT_NOTIFY, ReparentNotify)
			HANDLE(PROPERTY_NOTIFY, PropertyNotify)

			HANDLE(EXPOSE, Expose)
			default:
				{
					SEL extSel = 
						extensionSelectors[event->response_type  & ~0x80];
			
					if (extSel != (SEL)0)
					{
						[self performSelector: extSel
						           withObject: (id)event];
					}
					else
					{
						NSDebugLLog(@"XCBConnection",@"Don't yet know how to handle events of type %d (%hd)",
							event->response_type, event->sequence);
					}
				}
			case 0: {}
		}
		eventsHandled = YES;
		free(event);
	}
	return eventsHandled;
}
- (BOOL)handleReplies
{
	BOOL repliesHandled = NO;
	NSDebugLLog(@"XCBConnection",@"Handling replies");
	for(NSUInteger i=0 ; i<[replyHandlers count] ; i++)
	{
		NSArray *handler = [replyHandlers objectAtIndex: i];
		void *reply;
		xcb_generic_error_t *error;
		unsigned int sequenceNumber = 
			[[handler objectAtIndex: 0] unsignedIntValue];
		// FIXME: Handle errors.
		if (1 == xcb_poll_for_reply(connection, sequenceNumber, &reply, &error))
		{
			if (NULL == reply) 
			{
				// It is now safe to free the handler
				[replyHandlers removeObjectAtIndex: i];
				i--;
				continue;
			}
			NSDebugLLog(@"XCBConnection",@"Got reply for %d", sequenceNumber);
			if (error) 
			{
				NSLog(@"XCBConnection: ERROR for request seq %d: %d (response type %d)",
				error->sequence,
				error->error_code,
				error->response_type);
				if ([handler count] == 5)
				{
					id obj = [handler objectAtIndex: 1];
					id context = [handler objectAtIndex: 3];
					if ([context isEqual: [NSNull null]])
						context = nil;
					SEL errorSelector;
					[[handler objectAtIndex: 4] 
						getValue: &errorSelector];
					[obj performSelector: errorSelector
					          withObject: (id)error
					          withObject: (id)context];
				}
				free(error);
				continue;
			}
			id obj = [handler objectAtIndex: 1];
			SEL selector;
			[[handler objectAtIndex: 2] getValue: &selector];
			if ([handler count] == 3) 
			{
				[obj performSelector: selector withObject: (id)reply];
			}
			else
			{
				id context = [handler objectAtIndex: 3];
				[obj performSelector: selector
				          withObject: (id)reply
				          withObject: context];
			}
			free(reply);
			// Don't remove the handler just yet
			repliesHandled = YES;
		}
	}
	return repliesHandled;
}
- (void)setHandler: (id)anObject 
          forReply: (unsigned int)sequence
          selector: (SEL)aSelector
{
	NSNumber *key = [NSNumber numberWithUnsignedInt: sequence];
	NSDictionary *value = [NSArray arrayWithObjects:
		key,
		anObject,
		[NSValue valueWithBytes: &aSelector objCType: @encode(SEL)],
		nil];
	[replyHandlers addObject: value];
}
- (void)setHandler: (id)anObject
          forReply: (unsigned int)sequence
          selector: (SEL)aSelector
            object: (id)context
{
	NSNumber *key = [NSNumber numberWithUnsignedInt: sequence];
	NSDictionary *value = [NSArray arrayWithObjects:
		key,
		anObject,
		[NSValue valueWithBytes: &aSelector objCType: @encode(SEL)],
		context != nil ? context : (id)[NSNull null],
		nil];
	[replyHandlers addObject: value];
}
- (void)setHandler: (id)anObject
          forReply: (unsigned int)sequence
          selector: (SEL)aSelector
            object: (id)context
     errorSelector: (SEL)errorSelector
{
	NSNumber *key = [NSNumber numberWithUnsignedInt: sequence];
	NSDictionary *value = [NSArray arrayWithObjects:
		key,
		anObject,
		[NSValue valueWithBytes: &aSelector objCType: @encode(SEL)],
		context != nil ? context : (id)[NSNull null],
		[NSValue valueWithBytes: &errorSelector objCType: @encode(SEL)],
		nil];
	[replyHandlers addObject: value];
}

- (void)cancelHandlersForObject: (id)anObject
{
	// FIXME: This would be slow
	NSMutableIndexSet *removeObjects = [NSMutableIndexSet new];
	NSUInteger count = [replyHandlers count];
	for (NSUInteger i = 0; i < count; i++)
	{
		id replyHandler = [replyHandlers objectAtIndex:i];
		id handlerObject = [replyHandler objectAtIndex:2];
		if ([handlerObject isEqual: anObject])
			[removeObjects addIndex:i];

	}
	[replyHandlers removeObjectsAtIndexes: removeObjects];
	[removeObjects release];
}
- (void)handleError: (xcb_generic_error_t*)error
{
	XCBRaiseGenericErrorException(error, @"<unknown>", @"Unknown asynchronous request (from processing event loop)");
	NSDictionary *userInfo = [NSDictionary dictionaryWithObjectsAndKeys:
		[NSNumber numberWithUnsignedShort: error->sequence], @"Sequence",
		[NSNumber numberWithUnsignedChar: error->error_code], @"ErrorCode",
		[NSNumber numberWithUnsignedInteger: error->response_type], @"ResponseType",
		nil];
	NSException *e =
		[NSException exceptionWithName: XCBRequestErrorException
		                        reason: [NSString stringWithFormat: @"%s error (%s) - sequence %d", 
		                            xcb_event_get_error_label(error->error_code), 
		                            xcb_event_get_label(error->response_type), 
		                            error->sequence, nil]
		                      userInfo: userInfo];
	free(error);
	[e raise];
}
- (void)unregisterWindow: (XCBWindow*)aWindow
{
	NSDebugLLog(@"XCBConnection",@"Unregistering window: %@", aWindow);
	NSMapRemove(windows, (void*)(intptr_t)[aWindow xcbWindowId]);
}
- (void)registerWindow: (XCBWindow*)aWindow
{
	NSDebugLLog(@"XCBConnection",@"Registered window: %@", aWindow);
	NSMapInsert(windows, (void*)(intptr_t)[aWindow xcbWindowId], aWindow);
}
- (XCBWindow*)windowForXCBId: (xcb_window_t)anId;
{
	return NSMapGet(windows, (void*)(intptr_t)anId);
}
- (void)registerSelection: (XCBSelection*)selection
{
	NSDebugLLog(@"XCBConnection", @"Registered selection: %@", selection);
	NSMapInsert(selections, (void*)(intptr_t)[selection atom], selection);
}
- (void)unregisterSelection: (XCBSelection*)selection
{
	NSDebugLLog(@"XCBConnection", @"Unregistered selection: %@", selection);
	NSMapRemove(selections, (void*)(intptr_t)[selection atom]);
}
- (XCBSelection*)selectionWithAtom: (xcb_atom_t)atom
{
	return NSMapGet(selections, (void*)(intptr_t)atom);
}
- (void)setDelegate:(id) aDelegate
{
	delegate = aDelegate;
}
- (id)delegate 
{
	return delegate;
}
- (void)setSelector: (SEL)aSelector forXEvent: (uint8_t)anEvent
{
	extensionSelectors[anEvent] = aSelector;
}
- (NSArray*)screens
{
	return screens;
}
- (void)grab
{
	xcb_grab_server(connection);
	[self setNeedsFlush: YES];
}

- (void)ungrab
{
	xcb_ungrab_server(connection);
	[self setNeedsFlush: YES];
}
- (void)allowEvents: (xcb_allow_t)allow timestamp: (xcb_timestamp_t)time
{
	xcb_allow_events(connection, allow, time);
	// FIXME: Is flushing immediately too soon?
	[self flush];
}
- (uint8_t)grabPointerWithWindow: (XCBWindow*)grabWindow
                     ownerEvents: (BOOL)ownerEvents
                       eventMask: (uint16_t)eventMask
                     pointerMode: (uint8_t)pointerMode
                    keyboardMode: (uint8_t)keyboardMode
                       confineTo: (XCBWindow*)confineWindow
                          cursor: (xcb_cursor_t)cursor
                            time: (xcb_timestamp_t)time
{
	xcb_grab_pointer_cookie_t cookie = xcb_grab_pointer(connection,
		ownerEvents ? 1 : 0,
		[grabWindow xcbWindowId],
		eventMask,
		pointerMode,
		keyboardMode,
		[confineWindow xcbWindowId],
		cursor,
		time);
	xcb_grab_pointer_reply_t *reply = xcb_grab_pointer_reply(connection, cookie, NULL);
	uint8_t status = reply->status;
	free(reply);
	[self setNeedsFlush: YES];
	return status;
}
- (void)ungrabPointer: (xcb_timestamp_t)time
{
	xcb_ungrab_pointer(connection, time);
	[self setNeedsFlush: YES];
}
- (xcb_connection_t*) connection
{
	return connection;
}
- (void) dealloc
{
	xcb_disconnect(connection);
	[handle release];
	[replyHandlers release];
	[screens release];
	[super dealloc];
}
- (void)setNeedsFlush: (BOOL)shouldFlush
{
	needsFlush = shouldFlush;
}
- (BOOL)needsFlush
{
	return needsFlush;
}
- (xcb_timestamp_t)currentTime
{
	return currentTime;
}
- (void)flush
{
	xcb_flush(connection);
	needsFlush = NO;
}

@end

@implementation XCBConnection(Private)
- (void)eventsReady: (NSNotification*)notification
{
	// Poll while there is data left in the buffer
	while ([self handleEvents] || [self handleReplies]) {}
	// NSDebugLLog(@"XCBConnection", @"Finished handling events");
	if ([delegate respondsToSelector: @selector(finishedProcessingEvents:)])
		[delegate finishedProcessingEvents: self];
	if (needsFlush)
	{
		[self flush];
		needsFlush = NO;
	}
	[handle waitForDataInBackgroundAndNotify];
}
@end
