/**
 * Étoilé ProjectManager - XCBException.m
 *
 * Copyright (C) 2010 Christopher Armstrong <carmstrong@fastmail.com.au>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *
 **/

#import <XCBKit/XCBException.h>
#import <Foundation/NSValue.h>
#import <Foundation/NSException.h>
#import <Foundation/NSDictionary.h>

#include <xcb/xcb_event.h>

/**
  * An exception thrown containing the details of an error
  * generated by an X11 request.
  */
NSString *XCBRequestErrorException = @"XCBRequestErrorException";


void XCBRaiseGenericErrorException(xcb_generic_error_t* error, NSString* callName, NSString* description)
{
	NSDictionary *userInfo = [NSDictionary dictionaryWithObjectsAndKeys:
		[NSNumber numberWithUnsignedShort: error->sequence], @"Sequence",
		[NSNumber numberWithUnsignedChar: error->error_code], @"ErrorCode",
		[NSNumber numberWithUnsignedInteger: error->response_type], @"ResponseType",
		callName, @"CallName",
		description, @"CallDescription",
		nil];
	NSException *e =
		[NSException exceptionWithName: XCBRequestErrorException
		                        reason: [NSString stringWithFormat: @"%s error (%s) - sequence %d", 
		                            xcb_event_get_error_label(error->error_code), 
		                            xcb_event_get_label(error->response_type), 
		                            error->sequence, nil]
		                      userInfo: userInfo];
	free(error);
	[e raise];
}
